
import ipdlx.Strategy;

public class AMB extends Strategy {

  static double fading= 0.4;
  static double transfer= 0.5;
  static int history= 5;
  static int oppoffset= 1 << history;
  static int size= 1 << 2*history;
  static int mymask= ~(-1 << history);
  static int oppmask= (-1 << history) ^ (-1 << (2*history));
  static int allmask= ~(-1 << (2*history));

  int myrecent, opprecent;
  float[] experience;
  double mylast;
  boolean evercheated;

  public AMB() {
    super("AMB", "Ambitious", "Tries to predict the opponent's move based on past experience and chooses the move which seems to give most benefits.  The method is the same as for the MIR (Mirror) strategy.");
    reset();
  }

  public void reset() {
    int count;
    
    myrecent= mymask;
    opprecent= oppmask;
    mylast= COOPERATE;
    opponentMove= COOPERATE;
    lastResult= 0.5 * (DEFAULT_TEMPTATION + DEFAULT_SUCKER);
    experience= new float[size];
    for( count= 0; count< size; ++count )
      experience[count]= 0.5f;
    evercheated= false;
  }

  public double getMove()
  {
    updateExperience();
    myrecent= (myrecent << 1) & mymask;
    if( mylast==COOPERATE )
      ++myrecent;
    opprecent= (opprecent << 1) & oppmask;
    if( opponentMove==COOPERATE )
      opprecent += 1<<history;
    if( !evercheated || Math.random() <= experience[myrecent+opprecent] )
      mylast= COOPERATE;
    else
      mylast= DEFECT;
    return mylast;
  }

  protected void updateExperience()
  {
    int differing;
    double newexperience, weight;
    
    if( opponentMove==DEFECT )
      evercheated= true;
    weight= 1.0 - fading;
//    if( lastResult==DEFAULT_TEMPTATION )
//	  newexperience= 1.0;
//    else if( lastResult==DEFAULT_SUCKER )
//	  newexperience= 0.0;
//    else
//	  newexperience= 0.5;
    newexperience= (lastResult - DEFAULT_SUCKER)/(DEFAULT_TEMPTATION - DEFAULT_SUCKER);
    if( mylast==DEFECT )
      newexperience= 1.0 - newexperience;
    experience[myrecent+opprecent]= (float)(weight * newexperience +
			    (1.0 - weight) * experience[myrecent+opprecent]);
    for( differing= 1; differing< history; ++differing )
    {
      int[] diffbits= new int[differing];
      int count, testbit, bits, exp_ind;
      
      weight *= transfer;
      diffbits[0]= 1;
      for( count= 1; count< differing; ++count )
        diffbits[count]= diffbits[count-1] << 1;
      bits= ~(-1 << differing);
      while( true )
      {
	exp_ind= (myrecent + opprecent) ^ bits;
	experience[exp_ind]= (float)(weight * newexperience +
				(1.0 - weight) * experience[exp_ind]);
	for( count= differing-1; count>= 0; --count )
	  if( (diffbits[count] << (differing-count))< allmask )
	    break;
	if( count< 0 )
	  break;	//  no valid combination left
	bits ^= diffbits[count];
	diffbits[count] <<= 1;
	bits ^= diffbits[count];
	for( ++count; count< differing; ++count ) {
	  bits ^= diffbits[count];
	  diffbits[count]= diffbits[count-1] << 1;
	  bits ^= diffbits[count];
	}
      }
    }
  }
}


