
import ipdlx.Strategy;

public class MIR extends Strategy {

  static double fading= 0.4;
  static double transfer= 0.5;
  static int history= 5;
  static int oppoffset= 1 << history;
  static int size= 1 << 2*history;
  static int mymask= ~(-1 << history);
  static int oppmask= (-1 << history) ^ (-1 << (2*history));
  static int allmask= ~(-1 << (2*history));

  int myrecent, opprecent;
  float[] experience;
  double mylast;
  boolean evercheated;

  public MIR() {
    super("MIR", "Mirror", "Tries to mirror the opponent's moves, based on past experience.  The result of each move is stored in an array indexed by the previous moves.  This experience spills over to neighbouring array elements which represent a slightly different history.  To determine how to move, the algorithm compares a pseudo-random number with the value of the array element corresponding to the current move history.");
    reset();
  }

  public void reset() {
    int count;
    
    myrecent= mymask;
    opprecent= oppmask;
    mylast= COOPERATE;
    opponentMove= COOPERATE;
    experience= new float[size];
    for( count= 0; count< size; ++count )
      experience[count]= 0.5f;
    evercheated= false;
  }

  public double getMove()
  {
    updateExperience();
    myrecent= (myrecent << 1) & mymask;
    if( mylast==COOPERATE )
      ++myrecent;
    opprecent= (opprecent << 1) & oppmask;
    if( opponentMove==COOPERATE )
      opprecent += 1<<history;
    if( !evercheated || Math.random() <= experience[myrecent+opprecent] )
      mylast= COOPERATE;
    else
      mylast= DEFECT;
    return mylast;
  }

  protected void updateExperience()
  {
    int differing;
    double newexperience, weight;
    
    if( opponentMove==DEFECT )
      evercheated= true;
    weight= 1.0 - fading;
    newexperience= (opponentMove - DEFECT)/(COOPERATE - DEFECT);
    experience[myrecent+opprecent]= (float)(weight * newexperience +
			    (1.0 - weight) * experience[myrecent+opprecent]);
    for( differing= 1; differing< history; ++differing )
    {
      int[] diffbits= new int[differing];
      int count, testbit, bits, exp_ind;
      
      diffbits[0]= 1;
      weight *= transfer;
      for( count= 1; count< differing; ++count )
        diffbits[count]= diffbits[count-1] << 1;
      bits= ~(-1 << differing);
      while( true )
      {
	exp_ind= (myrecent + opprecent) ^ bits;
	experience[exp_ind]= (float)(weight * newexperience +
				(1.0 - weight) * experience[exp_ind]);
	for( count= differing-1; count>= 0; --count )
	  if( (diffbits[count] << (differing-count))< allmask )
	    break;
	if( count< 0 )
	  break;	//  no valid combination left
	bits ^= diffbits[count];
	diffbits[count] <<= 1;
	bits ^= diffbits[count];
	for( ++count; count< differing; ++count ) {
	  bits ^= diffbits[count];
	  diffbits[count]= diffbits[count-1] << 1;
	  bits ^= diffbits[count];
	}
      }
    }
  }
}


